@@ -1,4 +1,4 @@
-#!/usr/bin/python
+#!/usr/bin/python3
 # vim:set fileencoding=utf-8 et ts=4 sts=4 sw=4:
 #
 #   apt-listchanges - Show changelog entries between the installed versions
@@ -7,6 +7,7 @@
 #
 #   Copyright (C) 2000-2006  Matt Zimmerman  <mdz@debian.org>
 #   Copyright (C) 2006       Pierre Habouzit <madcoder@debian.org>
+#   Copyright (C) 2016       Robert Luberda  <robert@debian.org>
 #
 #   This program is free software; you can redistribute it and/or modify
 #   it under the terms of the GNU General Public License as published by
@@ -26,69 +27,62 @@
 
 import sys, os, os.path
 import apt_pkg
-import anydbm
-import commands
+import signal
+import subprocess
 import traceback
 
 sys.path += [os.path.dirname(sys.argv[0]) + '/apt-listchanges', '/usr/share/apt-listchanges']
-from ALChacks import *
-import apt_listchanges, DebianFiles, ALCConfig
+import ALCLog
+from ALChacks import _
+import apt_listchanges, DebianFiles, ALCApt, ALCConfig, ALCSeenDb
 
 def main(config):
     config.read('/etc/apt/listchanges.conf')
     debs = config.getopt(sys.argv)
 
+    if config.dump_seen:
+        ALCSeenDb.make_seen_db(config, True).dump()
+        sys.exit(0);
+
     apt_pkg.init_system()
 
     if config.apt_mode:
-        # stdin is a pipe
-        debs = apt_listchanges.read_apt_pipeline(config)
-
-        if config.debug:
-            sys.stderr.write("To list:\n")
-            for d in debs: sys.stderr.write("\t%s\n" % d)
-            sys.stderr.write("\n")
-
+        debs = ALCApt.AptPipeline(config).read()
         if not debs:
             sys.exit(0)
 
-    try:
-        # Give any forked processes (eg. lynx) a normal stdin;
-        # See Debian Bug #343423
-        tty = open('/dev/tty', 'r+')
-        os.close(0)
-        os.dup2(tty.fileno(), 0)
-        tty.close()
-    except:
-        pass
-
     # Force quiet (loggable) mode if not running interactively
     if not sys.stdout.isatty() and not config.quiet:
         config.quiet = 1
 
     try:
-        frontend = apt_listchanges.make_frontend(config.frontend, len(debs), config)
+        frontend = apt_listchanges.make_frontend(config, len(debs))
     except apt_listchanges.EUnknownFrontend:
-        sys.stderr.write(_("Unknown frontend: %s\n") % config.frontend)
+        ALCLog.error(_("Unknown frontend: %s") % config.frontend)
         sys.exit(1)
 
     if frontend == None:
         sys.exit(0)
 
+    if frontend.needs_tty_stdin() and not sys.stdin.isatty():
+         try:
+             # Give any forked processes (eg. lynx) a normal stdin;
+             # See Debian Bug #343423.  (Note: with $APT_HOOK_INFO_FD
+             # support introduced in version 3.2, stdin should point to
+             # a terminal already, so there should be no need to reopen it).
+             tty = open('/dev/tty', 'rb+', buffering=0)
+             os.close(0)
+             os.dup2(tty.fileno(), 0)
+             tty.close()
+         except Exception as ex:
+             ALCLog.warning(_("Cannot reopen /dev/tty for stdin: %s") % str(ex))
+
     if not config.show_all:
         status = DebianFiles.ControlParser()
         status.readfile('/var/lib/dpkg/status')
         status.makeindex('Package')
 
-    if config.save_seen:
-        try:
-            seen = anydbm.open(config.save_seen, 'c')
-            seen.has_key('foo%0')
-        except:
-            sys.stderr.write(_("database %s failed to load.\n") % config.save_seen)
-            sys.exit(1)
-        # Will replace seen after changes have actually been seen
-        seen_new = {}
+    seen_db = ALCSeenDb.make_seen_db(config)
 
     all_news = {}
     all_changelogs = {}
@@ -114,8 +108,8 @@ def main(config):
         fromversion = None
 
         if not config.show_all:
-            if config.save_seen and seen.has_key(srcpackage):
-                fromversion = seen[srcpackage]
+            if srcpackage in seen_db:
+                fromversion = seen_db[srcpackage]
             elif config.since:
                 fromversion = config.since
             else:
@@ -135,32 +129,29 @@ def main(config):
         #
         # This is why even if we've seen a package we may miss bits of
         # changelog in some odd cases
-        if found.has_key(srcpackage) and \
+        if srcpackage in found and \
                 apt_pkg.version_compare(srcversion, found[srcpackage]) <= 0:
             continue
 
         if not config.show_all and apt_pkg.version_compare(fromversion, srcversion) >= 0:
-            notes.append(_("%s: Version %s has already been seen") % (binpackage,
-                                                                      srcversion))
+            notes.append(_("%(pkg)s: Version %(version)s has already been seen")
+                           % {'pkg': binpackage, 'version': srcversion})
             continue
 
         (news, changelog, binnmu) = pkg.extract_changes(config.which, fromversion, config.reverse)
 
         if news or changelog or binnmu:
             found[srcpackage] = srcversion
+            seen_db[srcpackage] = srcversion
             if news:
                 all_news[srcpackage] = news
             if changelog:
                 all_changelogs[srcpackage] = changelog
             if binnmu:
                 all_binnmus[srcpackage] = binnmu
-            if config.save_seen:
-                seen_new[srcpackage] = srcversion
 
     frontend.progress_done()
-
-    if config.save_seen:
-        seen.close()
+    seen_db.close_db()
 
     # Merge binnmu entries with regular changelog entries.
     # Assumption: the binnmu version is greater than the last non-binnmu version.
@@ -170,12 +161,11 @@ def main(config):
         else:
             all_changelogs[srcpackage] = all_binnmus[srcpackage]
 
-    all_news = all_news.values()
-    all_changelogs = all_changelogs.values()
+    all_news = list(all_news.values())
+    all_changelogs = list(all_changelogs.values())
 
     for batch in (all_news, all_changelogs):
-        batch.sort(lambda a, b: -cmp(a.urgency, b.urgency) or
-                   cmp(a.package, b.package))
+        batch.sort(key=lambda x: (x.urgency, x.package))
 
     if config.headers:
         changes = ''
@@ -186,7 +176,7 @@ def main(config):
 
             package = rec.package
             header = _('News for %s') % package
-            if len(filter(lambda x: x != package, source_packages[package])) > 0:
+            if len([x for x in source_packages[package] if x != package]) > 0:
                 # Differing source and binary packages
                 header += ' (' + ' '.join(source_packages[package]) + ')'
             news += '--- ' + header + ' ---\n' + rec.changes
@@ -197,7 +187,7 @@ def main(config):
 
             package = rec.package
             header = _('Changes for %s') % package
-            if len(filter(lambda x: x != package, source_packages[package])) > 0:
+            if len([x for x in source_packages[package] if x != package]) > 0:
                 # Differing source and binary packages
                 header += ' (' + ' '.join(source_packages[package]) + ')'
             changes += '--- ' + header + ' ---\n' + rec.changes
@@ -219,39 +209,47 @@ def main(config):
     if news or changes:
         apt_listchanges.confirm_or_exit(config, frontend)
 
-        hostname = commands.getoutput('hostname')
+        hostname = subprocess.getoutput('hostname')
 
         if apt_listchanges.can_send_emails(config):
             if changes:
                 subject = _("apt-listchanges: changelogs for %s") % hostname
-                apt_listchanges.mail_changes(config.email_address, config.email_format,
-                                             changes, subject)
+                apt_listchanges.mail_changes(config, changes, subject)
 
             if news:
                 subject = _("apt-listchanges: news for %s") % hostname
-                apt_listchanges.mail_changes(config.email_address, config.email_format,
-                                            news, subject)
+                apt_listchanges.mail_changes(config, news, subject)
 
         # Write out seen db
-        if config.save_seen:
-            seen = anydbm.open(config.save_seen, 'c')
-            for (key, value) in seen_new.items():
-                seen[key] = value
-            seen.close()
+        seen_db.apply_changes()
 
     elif not config.apt_mode and not source_packages.keys():
-        sys.stderr.write("apt-listchanges: "
-                         + _("didn't find any valid .deb archives") + "\n")
+        ALCLog.error(_("Didn't find any valid .deb archives"))
         sys.exit(1)
 
+
+def _setup_signals():
+    def signal_handler(signum, frame):
+        ALCLog.error(_('Received signal %d, exiting') % signum)
+        sys.exit(apt_listchanges.BREAK_APT_EXIT_CODE)
+
+    for s in [ signal.SIGHUP, signal.SIGQUIT, signal.SIGTERM ]:
+        signal.signal(s, signal_handler)
+
 if __name__ == '__main__':
+    _setup_signals()
     config = ALCConfig.ALCConfig()
     try:
         main(config)
     except KeyboardInterrupt:
-        sys.exit(10)
+        sys.exit(apt_listchanges.BREAK_APT_EXIT_CODE)
+    except ALCApt.AptPipelineError as ex:
+        ALCLog.error(str(ex))
+        sys.exit(apt_listchanges.BREAK_APT_EXIT_CODE)
+    except ALCSeenDb.DbError as ex:
+        ALCLog.error(str(ex))
+        sys.exit(1)
     except Exception:
         traceback.print_exc()
-        apt_listchanges.confirm_or_exit(config, apt_listchanges.ttyconfirm())
+        apt_listchanges.confirm_or_exit(config, apt_listchanges.ttyconfirm(config))
         sys.exit(1)
-
