@@ -6,6 +6,7 @@
 #
 #   Copyright (C) 2000-2006  Matt Zimmerman  <mdz@debian.org>
 #   Copyright (C) 2006       Pierre Habouzit <madcoder@debian.org>
+#   Copyright (C) 2016       Robert Luberda  <robert@debian.org>
 #
 #   This program is free software; you can redistribute it and/or modify
 #   it under the terms of the GNU General Public License as published by
@@ -24,38 +25,42 @@
 #
 
 import re
-import sys, os, os.path
+import sys, os
 import tempfile
 import gzip
 import errno
 import glob
 import shutil
+import shlex
 import signal
 import subprocess
 
 import apt_pkg
-from ALChacks import *
+import ALCLog
+from ALChacks import _
+from functools import reduce
 
 # TODO:
 # indexed lookups by package at least, maybe by arbitrary field
 
-def numeric_urgency(u):
-    urgency_map = { 'low' : 1,
-                    'medium' : 2,
-                    'high' : 3,
-                    'emergency' : 4,
-                    'critical' : 4 }
+def _numeric_urgency(u):
+    urgency_map = { 'critical'  : 1,
+                    'emergency' : 1,
+                    'high'      : 2,
+                    'medium'    : 3,
+                    'low'       : 4 }
 
-    return urgency_map.get(u.lower(), 5)
+    return urgency_map.get(u.lower(), 0)
 
 
 class ControlStanza:
     source_version_re = re.compile('^\S+ \((?P<version>.*)\).*')
+    fields_to_read = [ 'Package', 'Source', 'Version', 'Architecture', 'Status' ]
 
-    def __init__(self, str):
+    def __init__(self, s):
         field = None
 
-        for line in str.split('\n'):
+        for line in s.split('\n'):
             if not line:
                 break
             if line[0] in (' ', '\t'):
@@ -63,7 +68,11 @@ class ControlStanza:
                     setattr(self, field, getattr(self, field) + '\n' + line)
             else:
                 field, value = line.split(':', 1)
-                setattr(self, field, value.lstrip())
+                if field in self.fields_to_read:
+                    setattr(self, field, value.lstrip())
+                else:
+                    field = None
+
 
     def source(self):
         return getattr(self, 'Source', self.Package).split(' ')[0]
@@ -103,15 +112,24 @@ class ControlParser:
             self.index[field][getattr(stanza, field)] = stanza
 
     def readfile(self, file):
-        self.stanzas += [ControlStanza(x) for x in open(file, 'r').read().split('\n\n') if x]
+        try:
+            self.stanzas += [ControlStanza(x) for x in open(file, 'r', encoding='utf-8', errors='replace').read().split('\n\n') if x]
+        except Exception as ex:
+            raise RuntimeError(_("Error processing '%(what)s': %(errmsg)s") %
+                                {'what': file, 'errmsg': str(ex)}) from ex
 
     def readdeb(self, deb):
-        fh = os.popen('dpkg-deb -f %s' % deb)
-        self.stanzas.append(ControlStanza(fh.read()))
+        try:
+            command = ['dpkg-deb', '-f', deb] + ControlStanza.fields_to_read
+            output = subprocess.check_output(command)
+            self.stanzas.append(ControlStanza(output.decode('utf-8', 'replace')))
+        except Exception as ex:
+            raise RuntimeError(_("Error processing '%(what)s': %(errmsg)s") %
+                                {'what': file, 'errmsg': str(ex)}) from ex
 
     def find(self, field, value):
-        if self.index.has_key(field):
-            if self.index[field].has_key(value):
+        if field in self.index:
+            if value in self.index[field]:
                 return self.index[field][value]
             else:
                 return None
@@ -121,9 +139,75 @@ class ControlParser:
                     return stanza
         return None
 
-class Package:
-    changelog_header = re.compile('^\S+ \((?P<version>.*)\) .*;.*urgency=(?P<urgency>\w+).*')
+class ChangelogParser:
+    _changelog_header = re.compile('^\S+ \((?P<version>.*)\) .*;.*urgency=(?P<urgency>\w+).*')
+    _changelog_header_ancient = re.compile('^(\S+ \(?\d.*\)|Old Changelog:|Changes|ChangeLog begins|Mon|Tue|Wed|Thu|Fri|Sat|Sun).*')
+    _changelog_header_emacs = re.compile('(;;\s*)?Local\s+variables.*', re.IGNORECASE)
 
+    def __init__(self):
+        self._entry = ''
+        self._entries = []
+
+    def parse(self, fd, since_version):
+        '''Parse changelog or news from the given file descriptor.
+        If since_version is specified, only return entries later
+        than the specified version.'''
+        urgency = _numeric_urgency('low')
+        is_debian_changelog = False
+        ancient = False
+
+        for line in fd.readlines():
+            line = line.decode('utf-8', 'replace')
+
+            if line.startswith(' ') or line == '\n':
+                self._add_line(line)
+
+            elif line.startswith('#'):
+                continue
+
+            else:
+                match = self._changelog_header.match(line) if not ancient else None
+                if match:
+                    if since_version and apt_pkg.version_compare(match.group('version'),
+                                                                 since_version) <= 0:
+                        break
+                    is_debian_changelog = True
+                    urgency = min(_numeric_urgency(match.group('urgency')),
+                                  urgency)
+                    self._save_entry(line)
+
+                elif self._changelog_header_ancient.match(line):
+                    if not is_debian_changelog: # probably upstream changelog in GNU format
+                        break
+                    ancient = True
+                    self._save_entry(line)
+
+                elif self._changelog_header_emacs.match(line):
+                    break
+
+                elif is_debian_changelog:
+                    self._add_line(line)
+
+                else:
+                    break
+
+        if not is_debian_changelog:
+            return None, urgency
+
+        self._save_entry(None)
+
+        return self._entries, urgency
+
+    def _add_line(self, line):
+        self._entry += line
+
+    def _save_entry(self, new_entry_header):
+        self._entry = self._entry.strip()
+        if self._entry != '':
+            self._entries += [ self._entry + '\n' ]
+        self._entry = new_entry_header
+
+class Package:
     def __init__(self, path):
         self.path = path
 
@@ -154,9 +238,9 @@ class Package:
             filenames.extend(changelog_filenames_binnmu)
             filenames.extend(changelog_filenames_native)
 
-        tempdir = self.extract_contents(filenames)
+        tempdir = self._extract_contents(filenames)
 
-        find_first = lambda acc, fname: acc or self.read_changelog(os.path.join(tempdir, fname), since_version, reverse)
+        find_first = lambda acc, fname: acc or self._read_changelog(os.path.join(tempdir, fname), since_version, reverse)
 
         news       = reduce(find_first, news_filenames, None)
         changelog  = reduce(find_first, changelog_filenames + changelog_filenames_native, None)
@@ -166,12 +250,12 @@ class Package:
 
         return (news, changelog, binnmu)
 
-    def extract_contents(self, filenames):
+    def _extract_contents(self, filenames):
         tempdir = tempfile.mkdtemp(prefix='apt-listchanges')
 
         extract_command = 'dpkg-deb --fsys-tarfile %s | tar xf - --wildcards -C %s %s 2>/dev/null' % (
-            self.path, tempdir,
-            ' '.join(["'" + x + "'" for x in filenames])
+            shlex.quote(self.path), shlex.quote(tempdir),
+            ' '.join([shlex.quote(x) for x in filenames])
         )
 
         # tar exits unsuccessfully if _any_ of the files we wanted
@@ -184,70 +268,50 @@ class Package:
 
         return tempdir
 
-    def read_changelog(self, filename, since_version, reverse=False):
+    def _open_changelog_file(self, filename):
         filenames = glob.glob(filename)
 
-        fd = None
         for filename in filenames:
             try:
                 if os.path.isdir(filename):
-                    print >> sys.stderr, _("Ignoring `%s' (seems to be a directory!)") % filename
+                    ALCLog.error(_("Ignoring `%s' (seems to be a directory!)") % filename)
                 elif filename.endswith('.gz'):
-                    fd = gzip.GzipFile(filename)
+                    return gzip.GzipFile(filename)
                 else:
-                    fd = open(filename)
+                    return open(filename, 'rb')
                 break
-            except IOError, e:
+            except IOError as e:
                 if e.errno == errno.ENOENT:
                     pass
                 else:
                     raise
+        return None
 
+    def _read_changelog(self, filename, since_version, reverse=False):
+        fd = self._open_changelog_file(filename)
         if not fd:
             return None
 
-        urgency = numeric_urgency('low')
-        entry = ''
-        entries = []
-        is_debian_changelog = 0
-        for line in fd.readlines():
-            match = self.changelog_header.match(line)
-            if match:
-                entries += [entry]
-                entry = ''
-                is_debian_changelog = 1
-                if since_version:
-                    if apt_pkg.version_compare(match.group('version'),
-                                              since_version) > 0:
-                        urgency = max(numeric_urgency(match.group('urgency')),
-                                      urgency)
-                    else:
-                        break
-            entry += line
+        entries, urgency = ChangelogParser().parse(fd, since_version)
 
-        if not is_debian_changelog:
-            return None
+        if not urgency:
+            return None # not a Debian changelog file
 
-        if entry != '':
-           entries += [entry]
+        if not entries:
+            # Valid Debian changelog, but all entries filtered out due
+            # to --since-version. Return empty string to prevent the caller
+            # from searching for another Debian changelog file.
+            return Changes(self.source, "", urgency)
 
         if reverse:
             entries.reverse()
-        changes = "".join(entries)
 
-        if changes and not changes.endswith('\n\n'):
-            changes += '\n'
+        changes = "\n".join(entries) + "\n"
 
         return Changes(self.source, changes, urgency)
 
     def _changelog_variations(self, filename):
-        formats = ['usr/doc/*/%s.gz',
-                   'usr/share/doc/*/%s.gz',
-                   'usr/doc/*/%s',
-                   'usr/share/doc/*/%s',
-                   './usr/doc/*/%s.gz',
-                   './usr/share/doc/*/%s.gz',
-                   './usr/doc/*/%s',
+        formats = ['./usr/share/doc/*/%s.gz',
                    './usr/share/doc/*/%s']
         return [x % filename for x in formats]
 
@@ -259,7 +323,7 @@ class Changes:
 
     def merge_binnmu(self, other, reverse = False):
         assert self.package == other.package
-        self.urgency = max(self.urgency, other.urgency)
+        self.urgency = min(self.urgency, other.urgency)
         # Assumption: binnmu has greater version than regular changelog
         if reverse:
             self.changes = self.changes + other.changes
